
/*****************************************************************************************
*
*  Vinhthuy Phan
*  ploc.h
*  06/25/97
*
*****************************************************************************************/

#include "pointloc.h"

/*****************************************************************************************
*
*  PointLocation: locate the triangle that contains a query point.  Assume the existence
*  of a Delaunay triangulation.
*
*  The algorithm is a 2D version of the "jump-and-walk" randomized algorithm invented by 
*  Ernst P. Mucke, Isaac Saias, and  Binhai Zhu, "Fast Randomized Point Location Without
*  Processing in Two- and Three-dimensional Delaunay Triangulations," Proceedings of the
*  12th Annual Symposium on Computational Geometry, 1996, p. 274-283.
*
*  The expected running time is O(n^(1/(d+1))) for point location in Delaunay triangulations
*  of n points, where d is the dimension; in our case the expected running time is
*  O(n^(1/3)).
*
*  The algorithm is simple. wwe start out by copying the triangulations into 
*  (local) global variables.  Then we choose k random edges, jump to the closest edge
*  from the query point, and start walking toward it.
*
*  Every walk from edge to edge is done in constant time.  When the line between the
*  first edge and the query point goes through a vertex of a triangle, we simply choose
*  another point close to this vertex and starting walking from there.  So the walk
*  may not be in a straight line.  However, by doing this we avoid the complication
*  involving floating point numbers when we try to compute which triangle is 
*  that contains the vertex through which the line crosses.  This convention guarantees
*  that starting from the nearest random point to the query the line always enters an
*  edge of a triangle.
*
*  The number of random edges is currently chosen to be n^(1/3), n is the number of
*  edges.
*
*	if iEdgeID >= 0 then that is where the algorithm will start walking from
*
*****************************************************************************************/
long PointLocation(POINT_TYPE QP, void* pPET[], long aiNum[], long iEdgeID)
{
    long iIntersectedTri = -1; /** stores the tri intersected by the line most recently **/
    STACK_NODE_PTR spEdgeIDs = NULL;
    POINT_TYPE sFirstPoint;  /** a chosen point to start walking to QP **/
    long *iRandomEdges, first_edge, num_of_random_edges; 
    int     iC=0;
    
    /************* precomputations **************/
    SetupGlobal(pPET, aiNum);
    NUM_OF_INTERSECTED_EDGES = 0;
	if (iEdgeID < 0) {
    	num_of_random_edges = (long) CubeRoot((float) NUM_OF_VET[0]);
		if (num_of_random_edges < 2)	num_of_random_edges = 2;
    	iRandomEdges = calloc( num_of_random_edges, sizeof(long));
    	GetKRandomEdges( iRandomEdges, num_of_random_edges ); 
	}
	else {
    	num_of_random_edges = 1;
    	iRandomEdges = calloc( num_of_random_edges, sizeof(long));
		*iRandomEdges = iEdgeID;
	}
    first_edge = ClosestEdgeFromQP( QP, iRandomEdges, num_of_random_edges, &sFirstPoint );
    StackPush( &spEdgeIDs, first_edge );
    NUM_OF_INTERSECTED_EDGES = 1;
    iIntersectedTri = GetTriangleContainingEdge( spEdgeIDs->ID, iIntersectedTri );
    if(iIntersectedTri == -1) return -1;

    /************* start walking **************/
    while( ! InsideTri( QP, pgTRIANGLES[iIntersectedTri] )) {
		FindEdge_UpdatePoint( iIntersectedTri, &spEdgeIDs, &sFirstPoint, QP );
		iIntersectedTri = GetTriangleContainingEdge( spEdgeIDs->ID, iIntersectedTri );
		if(iIntersectedTri == -1 || iC == 1000){
		    return -1;
		}
		iC++;		
    }
/**
    StackPrint(spEdgeIDs);
 **/
 	free(iRandomEdges);

    return iIntersectedTri;
    
}
/***************************************************************************************
* 
*  Called in PointLocation.  This procedure stores the triangulation given in pPET
*  and other info in aiNum into (global) local variables
*
***************************************************************************************/
void SetupGlobal(void* pPET[3], long aiNum[])
{
    pgVERTICES = (VERTEX_TYPE*) pPET[0];
    pgEDGES = (EDGE_TYPE*) pPET[1];
    pgTRIANGLES = (TRIANGLE_TYPE*) pPET[2];
    NUM_OF_VET[0] = aiNum[0];   /** [0]: # of vertices **/
    NUM_OF_VET[1] = aiNum[1];   /** [1]: # of edges **/
    NUM_OF_VET[2] = aiNum[2];   /** [2]: # of triangles **/
}

/*****************************************************************************************
*  Called in PointLocation.
*  Given the most recent edge (on top of the stack) crossed over by the line and 
*  the triangle containing it, this function push onto the stack the next edge of 
*  the triangle crossed over by the line.  If the next is crossed over through a vertex
*  *sFirstPoint is updated:  we start walking from a new point (in this case, the midpoint
*  of the edge) to guarantee that the line never crosses a vertex.
*  
*  Note that this new point can be arbitrarily close to the crossed-over vertex
*  so this should not greatly effect the performance of the walk.
******************************************************************************************/

void FindEdge_UpdatePoint(long iTri, STACK_NODE_PTR *spEdgeIDs,
			  POINT_TYPE *sFirstPoint, POINT_TYPE sLastPoint)
{
    EDGE_TYPE sTmpEdge;
    POINT_TYPE p1, p2;
    int index;

    for(index=0; index<3; index++){
	GET_EDGE( pgTRIANGLES[iTri].iEdge[index], &sTmpEdge );
	GET_POINT( sTmpEdge.iPoint[0], &p1 );
	GET_POINT( sTmpEdge.iPoint[1], &p2 );
	if( Intersect_Segments2D( p1, p2, *sFirstPoint, sLastPoint ) &&
	   ((*spEdgeIDs)->ID != pgTRIANGLES[iTri].iEdge[index]) ){
	    if( Between(*sFirstPoint, sLastPoint, p1) || Between(*sFirstPoint, sLastPoint, p2) ){
		FindMidPoint(p1, p2, sFirstPoint);
	    }
	    NUM_OF_INTERSECTED_EDGES++;
	    StackPush( spEdgeIDs, pgTRIANGLES[iTri].iEdge[index] );
	    return;
	}
    }
}

/****************************************************************************************
*  Called in PointLocation.  This procedure load k random edges into *iEdge to
*  start the random walk
*****************************************************************************************/
void GetKRandomEdges(long *iEdge, long k)
{
    long index, i;
    srand(time(NULL)); 

    for(i=0; i<k; i++)	{
	index = rand() % NUM_OF_VET[1];
	iEdge[i] = index;
    }
}
/*****************************************************************************************
*  Called in PointLocation.
*  Every edge (except those on the boundary) has exactly two pointers to two triangles
*  to which it belongs.  Given an edge and a triangle containing it, the procedure
*  returns the other triangle containing the edge.
*
*  If the edge is on the boundary, i.e. NUM_OF_INTERSECTED_EDGES = 1, we simply return
*  the only one triangle containing the edge.  It is impossible that in the middle 
*  of a walk, a boundary edge is encountered.  This happens only when THE QUERY POINT 
*  IS OUTSIDE OF THE TRIANGULATION.  If this  happens -1 is returned.
*
******************************************************************************************/
long GetTriangleContainingEdge(long AnEdge, long TheTri )
{
    EDGE_TYPE sIntersectEdge;
  
    GET_EDGE( AnEdge, &sIntersectEdge );

    if( ! EdgeIsOnBorder( sIntersectEdge )){      
	if( sIntersectEdge.iTri[0] != TheTri ){
	    return sIntersectEdge.iTri[0];
	}
	else{
	    return sIntersectEdge.iTri[1];
	}
    }
    else if( NUM_OF_INTERSECTED_EDGES == 1 ){
	if( sIntersectEdge.iTri[0] == -1 ){  /** iTri[0] is nonexistent **/
	    return sIntersectEdge.iTri[1];
	}
	else {
	    return sIntersectEdge.iTri[0];
	}
    }
    else {  

/**
	printf("\nIn GetTriangleContainingEdge; edge is %li\nQP must be outside the polygon.\n", AnEdge);
	Pause();
**/

	return -1;
    }
}

/*****************************************************************************************
*
*  Checks to see if an edge is on the boundary.  It is if, by convention, one of the
*  triangle containing it has -1 as the id.
*
*****************************************************************************************/
bool EdgeIsOnBorder( EDGE_TYPE sEdge )
{
    if( ( sEdge.iTri[0] == -1 ) || ( sEdge.iTri[1] == -1 ) ){
	return 1;
    }
    else {
	return 0;
    }
}
/*****************************************************************************************
*
*  Called in PointLocation.
*  The procedure computes the nearest edge from the query point QP.  The the midpoint
*  of this edge is assigned to *sFirstPoint, the point from which the walk is started.
*
******************************************************************************************/
long ClosestEdgeFromQP(POINT_TYPE QP, long *iEdge, long size, POINT_TYPE *sFirstPoint)
{
    float distance, closest;
    long index, icurrent_edge;
    POINT_TYPE midpoint, p1, p2;
    EDGE_TYPE sEdge;

    for(index=0; index<size; index++){
	GET_EDGE( iEdge[index], &sEdge );
	GET_POINT( sEdge.iPoint[0], &p1 );
	GET_POINT( sEdge.iPoint[1], &p2 );
	FindMidPoint( p1, p2, &midpoint);
	distance = Macro_Dist( QP, midpoint );
	if( index== 0 )
	    {
		closest = distance;
		icurrent_edge = iEdge[index];
		sFirstPoint->fx = midpoint.fx;
		sFirstPoint->fy = midpoint.fy;
		sFirstPoint->fz = midpoint.fz;
	    }
	else if( (distance <= closest) )
	    {
		closest = distance;
		icurrent_edge = iEdge[index];
		sFirstPoint->fx = midpoint.fx;
		sFirstPoint->fy = midpoint.fy;
		sFirstPoint->fz = midpoint.fz;
	    }
    }

    return icurrent_edge;
}

/***************************************************************************************/
void FindMidPoint( POINT_TYPE p1, POINT_TYPE p2, POINT_TYPE *midpoint )
{
    midpoint->fx = ( p1.fx + p2.fx ) / (float) 2.0;
    midpoint->fy = ( p1.fy + p2.fy ) / (float) 2.0;
    midpoint->fz = ( p1.fz + p2.fz ) / (float) 2.0;
}
/***************************************************************************************/
void GET_POINT( long PointId, POINT_TYPE *Point )
{
    if((PointId < 0) || (PointId > NUM_OF_VET[0])){
	printf("\nIn GET_POINT: %li, illegal point.\n", PointId);
	exit(1);
    }
    Point->fx = pgVERTICES[PointId].fx;
    Point->fy = pgVERTICES[PointId].fy;
    Point->fz = pgVERTICES[PointId].fz;
}

void GET_EDGE( long EdgeId, EDGE_TYPE *Edge )
{
    if((EdgeId < 0) || (EdgeId > NUM_OF_VET[1])){
	printf("\nIn GET_EDGE: %li, illegal edge.\n", EdgeId);
	exit(1);
    }
    Edge->iPoint[0] = pgEDGES[EdgeId].iPoint[0];
    Edge->iPoint[1] = pgEDGES[EdgeId].iPoint[1];
    Edge->iTri[0] = pgEDGES[EdgeId].iTri[0];
    Edge->iTri[1] = pgEDGES[EdgeId].iTri[1];

}

void GET_TRIANGLE( long TriId, TRIANGLE_TYPE *Tri )
{
    if((TriId < 0) || (TriId >= NUM_OF_VET[2])){
	printf("\nIn GET_TRIANGLE: %li, illegal triangle.\n", TriId);
	exit(1);
    }
    Tri->iPoint[0] = pgTRIANGLES[TriId].iPoint[0];
    Tri->iPoint[1] = pgTRIANGLES[TriId].iPoint[1];
    Tri->iPoint[2] = pgTRIANGLES[TriId].iPoint[2];
    Tri->iEdge[0] = pgTRIANGLES[TriId].iEdge[0];
    Tri->iEdge[1] = pgTRIANGLES[TriId].iEdge[1];
    Tri->iEdge[2] = pgTRIANGLES[TriId].iEdge[2];
}

/*****************************************************************************************
*
*  Stack routines used to store primarily the edges crossed over by the line between
*  the first and query point.
*
*****************************************************************************************/
void StackPush( STACK_NODE_PTR *topPtr, long id )
{
    STACK_NODE_PTR newPtr;
    newPtr = malloc( sizeof (StackNode) );
    if( newPtr != NULL ){
	newPtr->ID = id;
	newPtr->next = *topPtr;
	*topPtr = newPtr;
    }
    else{
	printf("%d not inserted. Memory unavailable.\n", id);
    }
}
void StackPop( STACK_NODE_PTR *topPtr, long *id )
{
    STACK_NODE_PTR tmpPtr;
    if( *topPtr != NULL ){
	tmpPtr = *topPtr;
	*id = (*topPtr)->ID;
	*topPtr = (*topPtr)->next;
	free(tmpPtr);
    }
    else{
	printf("stack is empty; can't pop.\n");
    }
}
void StackPrint( STACK_NODE_PTR currentPtr )
{
    printf("Stack content is: \n");    
    while( currentPtr != NULL ){
	printf("%d\n", currentPtr->ID);
	currentPtr = currentPtr->next;
    }    
    printf("End of Stack\n");
}

/*****************************************************************************************
*
*  Util. functions
*
*****************************************************************************************/

float CubeRoot(float x)
{
    return (float) pow(x, 0.333333);
}

float Log2(float x)
{
    return (float) (log10(x) / log10(2.0));
}

void Pause()
{
  char c;
  printf(".........Pausing.......");
  c = getchar();
}
/*****************************************************************************************/

