/***************************************************************************************
		This is the PREDICT program.  It reads in:

- bounding region info
- obstacle info
- transmitter info
- query point locations
- user info

	and produces:

- a constrained delaunay triangulation of the obstacles (for the prediction models)
			(using TRIANGLE from Jack Snoeyink)
- a delaunay triangulation of the query points (for the query module)

	and then calls the user selected prediction method to be used (a user can define
  a different method for each transmitter).
***************************************************************************************/

/***************************************************************************************
  Variable labeling Conventions:
    i - integer (long or short)      All variable names will be prefixed by their type
    p - pointer
    f - float  (float or double)
    a - array
    s - structure
    u - union
	g - global
***************************************************************************************/
#ifndef	 PREDICT
#define PREDICT

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define			MAX_POINTS				10000
#define 		MAX_TRANS_POINTS 		1500
#define         MAX_EDGES               20000

#define Macro_Sign(v) ((v > .000001) ? (1) : ((v < -.000001) ? (-1) : (0)))

#define Macro_PointsEqual2D(s1, s2)	(s1.fx == s2.fx && s1.fy == s2.fy)
#define Macro_PointsEqual3D(s1, s2) (s1.fx == s2.fx && s1.fy == s2.fy && s1.fz == s2.fz)
#define Macro_Dist(sP1, sP2)       	((sP1.fx - sP2.fx)*(sP1.fx - sP2.fx) + (sP1.fy - sP2.fy)*(sP1.fy - sP2.fy) + (sP1.fz - sP2.fz)*(sP1.fz - sP2.fz))
#define Macro_SqrtDist(sP1, sP2)   	((float) sqrt(((sP1.fx - sP2.fx)*(sP1.fx - sP2.fx) + (sP1.fy - sP2.fy)*(sP1.fy - sP2.fy) + (sP1.fz - sP2.fz)*(sP1.fz - sP2.fz))))

#define Macro_DotProduct(sV1, sV2)	(sV1.i * sV2.i + sV1.j * sV2.j + sV1.k * sV2.k)

#define Macro_NotZero(f)			(f > .0001 || f < -.0001)
#define Macro_NotZeroErr(f, e)		(f > e || f < -1 * e)
#define Macro_Zero(f)				(f < .0001 && f > -.0001)
#define Macro_ZeroErr(f, e)			(f < e && f > -1 * e)
#define	Macro_FloatEqual(f1, f2)	(f1 < f2 + (float) .0001 && f1 > f2 - (float) .0001)

#define bool int

typedef struct 
{
  float fx,fy,fz;
} VERTEX_TYPE;

typedef VERTEX_TYPE		POINT_TYPE;

typedef		char	BYTE;

typedef	struct {
	float i, j, k;	
} UNIT_VECTOR;
typedef UNIT_VECTOR		VECTOR_TYPE;

typedef struct {
	UNIT_VECTOR		sN;	/* normal to the plane (point normal form) */
	VERTEX_TYPE		sP;
} PLANE_TYPE;

typedef struct {
	UNIT_VECTOR		sV;	/* a unit vector in the direction of the line */
	VERTEX_TYPE		sP;
} PARA_LINE_TYPE;
	
typedef struct  
{
  POINT_TYPE    sLoc;
  float         fFreq_LRange, fFreq_HRange;
  short		    iModel_Type;
  float 	    fMax_Power;
  short 		iOmni_directional;
  float         fAng1, fAng2;
  long*			piQPoints;		/* points to a list of query points that might be able */
  long 			iNum_Points;				/* to 'hear' this transmitter */
  short         iTrans_ID;
  short			iPolarization;	/* 0 vertical, 1 horizontal */
  		/* ID of the triangle from gpObst_Tri that contains this transmitter */
  long			iTriID;	
} TRANS_TYPE; 

typedef struct 
{
	short			iNumVert;	
	long*			piVert;
  	short			iWallType, iThickness;
} WALL_TYPE;


typedef struct SIGNAL_TYPE* SIGNAL_TYPEptr;
typedef struct SIGNAL_TYPE 
{
  float	      		fPath_Loss, fDelay_Spread;
  float				fShortPath, fShortStr;
  short             iTrans_ID;
  SIGNAL_TYPEptr   	psNext;
} SIGNAL_TYPE;

typedef struct 
{
  POINT_TYPE        sLoc;
  short             iBest_Trans_ID;
  float	     		fBest_Path_Loss;
  short				iNum_Trans_Can_Hear;
  SIGNAL_TYPE* 		psSignal_Data;
} QP_TYPE;

typedef struct 
{
  long 		iPoint[2];
  long 		iTri[2];
  short		iWallType;	 /* 0 - created during triangulation.  > 0, a wall type */
  short		iThickness;
} EDGE_TYPE;

typedef struct 
{
  long iPoint[3];
  long iEdge[3];
} TRIANGLE_TYPE;
	
	/* for storing which query points are contained within a triangle */
typedef struct {
	short	iNum;
	long*	piTriQP;
} TRI_QP_TYPE;

/***************************************************************************************
                           Global Variables

  gpsPoints_Tri[3]: an array that points to the triangle/edge/point data structure for the
        query points
  gpsObst_Tri[3]: an array that points to the triangle/edge/point data structure for the
        obstacles
  gpsObst_Walls: a pointer to a list of walls for all the obstacles
  gpsTrans_Data: points to the array of transmitter information
  giNum_Trans: number of transmitters
  gaiNum_Obst[4]:  aiNum_Obst[0] - number of vertices
                   aiNum_Obst[1] - number of edges
                   aiNum_Obst[2] - number of triangles
		  		   aiNum_Obst[3] - Number of Obstacles
  aiNum_QPoints[3]: aiNum_QPoints[0] - number of points
                    aiNum_QPoints[1] - number of edges
		    		aiNum_QPoints[2] - number of triangles
***************************************************************************************/
TRANS_TYPE*       	gpsTrans_Data;
short              	giNum_Trans;
QP_TYPE*			gpsQPoints;
long				giNumQP;
POINT_TYPE*			gpsTerrain;		/* points to terrain heigh field data */
long				giNumTerr;		/* number of points in the height field */
PLANE_TYPE*			gpsTerrPlanes;	/* planes match triangle's planes based on height 
										field.  Accessed by Triangle IDs. */
TRI_QP_TYPE*		gpsTriQP;

		/* Global Trianglulation variables */
void*               gpQP_Tri[3];
void*				gpObst_Tri[3];
long               	gaiNumQP[3];
long               	gaiNum_Obst[3];
VERTEX_TYPE*		gpsVert;		/* here for convienience and use in macros */		
EDGE_TYPE*			gpsEdge;
TRIANGLE_TYPE*		gpsTri;

	/* Global building storage variables (walls and vertices) */
WALL_TYPE*        	gpsObst_Walls;
VERTEX_TYPE*		gpsObst_Vertices;
long				giNumObstVert, giNumObstWalls;

	/* Global user definable parameters */
short              	gi3DOn, giObst;             

	/* global variables set by command line arguments */
short				giCL_Initialize;
short				giCL_recursion_level;
short				giCL_max_events;
double				gfCL_beta_threshold;
short				giCL_Split;
short				giCL_RaySave;
short				giCL_Geomview;
short				giCL_Debug;
short				giCL_Stats;
short				giCL_InBuilding;
short				giCL_Terrain;
short				giCL_SavePaths;
float				gfCL_Split;
FILE*				gfp_Debug;
FILE*				gfp_Paths;

float 			gfC;	/* contains the speed the radio signal will propagate */
										/* through the medium (air) */
double			gfFree; /* for storing a Free Space 'constant' to speed 
															calculations up */
float			gfFloorHeight;		/* height of each floor in InBuiling scenario */

	/* upper x and y coordinates of the bounding rectangle */
POINT_TYPE			gasBR[2];

char               	cc;				   /* dummy variable for debugging purposes */

	/* in Predict.c.  for storing the results of the modeling method */
void Add_Signal_Data(QP_TYPE* psQP, TRANS_TYPE sT, float fPower, float fShortPath, float fShortStr, float fLongPath, float fLongStr);

	/* modeling methods */
short Ray_Method(short iTransID, short iMaxReflect, double fBeta, short iPropForm, short iReflTransForm);

	/******* LOAD DATA FUNCTIONS *******/
VERTEX_TYPE* Load_QPoints_To_Predict();
WALL_TYPE* Load_Obstacles();
TRANS_TYPE* Load_Transmitter_Info();
void Load_Data();
void Get_Assumptions();

	/******** Data Structures Functions *******/
void Initialize_Data_Structures();
void Save_TriDS(char* sFileName, void* PET[], long iNum[]);
void Load_TriDS(char* sFileName, void* apPET[], long* piNumTri);
void Save_Query_Points(QP_TYPE* pQPoints, void* asPET[], long aiNum[]);
QP_TYPE* Load_Query_Points(void* asPET[], long aiNum[]);
void Compute_Trans_QPoints();
long OtherTri(long iPrevTri, long iEdgeID);
long Other_Edge(long iTriID, long iEdgeID1, long iEdgeID2);
POINT_TYPE Common_Edge_Vertex(long iEdgeID1, long iEdgeID2);
PLANE_TYPE* Compute_Terrain_Tri_Planes();
long Tri_Edge_Segment_Inter(long iTriID, POINT_TYPE sP1, POINT_TYPE sP2);
void Assign_QP_to_Triangles();

	/******** Geomview output routines ********/
void Geomview_Out(char* s);

	/******** Error Checking/Handling Routines ******/
void ERROR_Message(char* s1, char* s2);
void ERROR_Check_For_Duplicate_Vertices(VERTEX_TYPE* pVert, long iNumVert, char* s);
void ERROR_Check_Verts_Planar(long* piWallVert, short iNumVert);
void Check_Tri_Data(void* apTri[], long* piNum_Tri, short QUERY);
void Print_Triangle(long iTri);
void Print_Point(POINT_TYPE sP);
void Print_Vector(VECTOR_TYPE sV);
void fPrint_Point(FILE* fp, POINT_TYPE sP);
void fPrint_Vector(FILE* fp, VECTOR_TYPE sV);
void Print_Plane(PLANE_TYPE sPl);
void fPrint_Plane(FILE* fp, PLANE_TYPE sPl); 
void CloseDF(); void SaveDF(); void OpenDF();

	/******** Utility Routines *********/
char* Parse_Value(char c[], char s[]);
float Area2( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC );
bool XOR( bool x, bool y );
bool Left( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC );
bool Between( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC );
bool InsideTri( POINT_TYPE sP, TRIANGLE_TYPE sTri );
bool Inside_Tri( POINT_TYPE sP, POINT_TYPE p1, POINT_TYPE p2, POINT_TYPE p3);
bool Collinear( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC );
bool LieOn( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD );
bool Intersect( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD );
bool Intersect_Segments2D( POINT_TYPE sA, POINT_TYPE sB, POINT_TYPE sC, POINT_TYPE sD );
POINT_TYPE ComputeCentroid3D(POINT_TYPE sTriP[]);
UNIT_VECTOR Normalize_Vector(VECTOR_TYPE sV);
PARA_LINE_TYPE	Line_Parametric_Equation3D(VERTEX_TYPE sP1, VERTEX_TYPE sP2);
POINT_TYPE ParaLine_Calc_Point(VECTOR_TYPE sU, POINT_TYPE sO, float t);
POINT_TYPE Nearest_Point_On_Plane(PLANE_TYPE sPlane, POINT_TYPE sP);
float Point_To_Plane_DistancePM(PLANE_TYPE sPlane, POINT_TYPE sP);
float Point_To_Plane_DistancePM2(UNIT_VECTOR sN, POINT_TYPE sP1, POINT_TYPE sP2);
short Point_On_Plane(PLANE_TYPE sPlane, POINT_TYPE sP);
PLANE_TYPE Plane_Point_Normal_Plus(POINT_TYPE sP1, POINT_TYPE sP2, POINT_TYPE sP3, POINT_TYPE sPdir);
UNIT_VECTOR	Plane_Point_Normal_Minus(POINT_TYPE sP1, POINT_TYPE sP2, POINT_TYPE sP3, POINT_TYPE sPdir);
UNIT_VECTOR	Plane_Point_Normal(POINT_TYPE sP1, POINT_TYPE sP2, POINT_TYPE sP3);
POINT_TYPE	Line_Plane_Intersection(PARA_LINE_TYPE sLine, PLANE_TYPE sPlane);
UNIT_VECTOR Reflection_Vector(VECTOR_TYPE sV, PLANE_TYPE sPlane);
float Point_To_Plane_Distance(PLANE_TYPE sPlane, POINT_TYPE sP);
VECTOR_TYPE Vector_MultBy_Scaler(VECTOR_TYPE sV, float v);
POINT_TYPE MidPoint(POINT_TYPE p1, POINT_TYPE p2);
POINT_TYPE Calc_Intersection_Of_Segments2D(POINT_TYPE sM1, POINT_TYPE sM2, POINT_TYPE sN1, POINT_TYPE sN2, short* piFlag);
short VectorsEqual3D(VECTOR_TYPE sV1, VECTOR_TYPE sV2, float e);
int MyBinSearch(float fV, float fArr[], int iN, short iAscend);
POINT_TYPE Rotate_CosXYZ(POINT_TYPE sP, float fCosTheta, float fSinTheta, short iRot);
short PointsEqual3D(POINT_TYPE sP1, POINT_TYPE sP2, float e);

	/******** from PointLoc.c ********/
long PointLocation(POINT_TYPE QP, void* pPET[], long aiNum[], long iEdgeID);

	/****** RayInit.c ********/
int Initialize_Ray(int nlvls, VERTEX_TYPE (**Parr)[3]);	

#endif
