#include "predict.h"
#include "formulas.h"

#ifndef RAY_MODEL
#define RAY_MODEL

/*  Questions to for research:

1) 	2D with Rayleigh(?) fading vs. 3D modeling
2)  Coefficient effects on path-loss calculations

*/
/*
	Coding Questions:
1) fix/check problem where cone vector is parallel to the plane it is being intersected with
2) 
3) streamline the Ray-init function (documented how in the function)
4) better 3D above-below segment testing
5) command line arguments
*/

/***************************************************************************************
  Variable labeling Conventions:
    i - integer (long or short)      All variable names will be prefixed by their type
    p - pointer
    f - float  (float or double)
    a - array
    s - structure
    u - union
	g - global

	* apTri_Qp is a pointer to an array of long points that are used to store which
		query points are contained in a triangle.  The array is accessed by use of
		the chosen triangle's ID #.  This number determines the offset into the array.
		IDs start at 0.  The pointer at the location either points to a list of
		long integers where the first entry is the number of entries minus 1, or
		is NULL in which case this triangle contains no query points.

	* asQP_Store is an array of type QP_RAY_STORE_TYPE.  This array is used to store
		the signal strength and some other characteristics at the query point.
***************************************************************************************/

/*	
	We will use floating point numbers since there will be relatively few tests
  and exact accuracy is not necessary.  For example, saying a ray passes to the right
  of a triangle vertex rather than the left is not a problem since it will have no
  effect, or negligable effect, to the overall path-loss calculations.
  	The special cases will have to be handled whether we use floating or rational
  arithematic as well.
*/

#define		NUM_WALL_TYPES		10
#define 	MAX_EVENTS			4
#define 	MEM_PATHS			20
#define		MAX_NEW_TRI			500

	/* used during creation of new rays */
typedef struct {
	POINT_TYPE	asP[4];		/* asP[3] is the centroid of the tirangle */
	long		iEdgeID;	/* id of edgeplane the triangle points are incident upon */
								/* -2 if ground plane, -3 if ceiling */
	short		iPlane;		/* Plane the triangle rests upon - internal */
	short		iActive;	
} RAY_TRI_TYPE;

typedef struct {
	PLANE_TYPE	sP[5];
} TRAP_TYPE;

typedef struct {
	UNIT_VECTOR	sRayVect;
	UNIT_VECTOR	asTriVect[3];	/* for storing the 3 triangle line equations */
	POINT_TYPE	asTriP[3];		/* the 3 points that were used to define the ray-cone */
	float		fAtten;			/*	attenuation due to reflections and transmissions */
	VERTEX_TYPE	sVirtTransLoc;	/* virtual transmitter location */
	VERTEX_TYPE sLKP;			/* ray's last known position */
	long		iTriID;			/* a ray 'knows' which triangle it is in*/
	long		iEdgeID;		/* a ray 'knows' which edge it just intersected */
	char		cNumEvent;		/* number of reflections/transmissions/diffractions */
			/* stores the reflection/transmission history  */
	long		aiPath[MAX_EVENTS];	
	char		cPrevious;		/* '0' - nothing '1' - wall reflection '2' - 'wall transmission' */
								/* '3' - ground reflection '4' - 'ceiling reflection' */
								/* '5' - 'diffracted ray' */
	char		acEvent[MAX_EVENTS];
} RAY_TYPE;	

typedef struct {
		short 	iTransID;
		long	iNumQP;
		long*	piQP;

		char	cMaxEvents;
		double 	fBeta;
		short 	iAllowedEvents;
		short 	iPropForm; 
		short 	iReflTransForm;
		float	fCeiling;
		short	iNumRecurse;
		short   iPolarization;
} RAY_SETTINGS;

typedef struct QP_RAY_STORE 	*QP_RAY_STOREptr;
typedef struct QP_RAY_STORE {
	long			iQP_ID;
	float			fSignalPow;
		/* the strength of the strongest ray to reach this point and its distance traveled */
	float			fMaxPow, fMaxDist;	
		/* the strength and distance of the shortest ray to reach this point */
	float			fShortPow, fShortDist;
		/* the strength and distance of the longest ray to reach this point */
	float			fLongPow, fLongDist;
} QP_RAY_STORE_TYPE;

typedef struct 
{
	char		cNumEvent, acEvent[4];
	long		aiPath[4];
	POINT_TYPE	sLKP;
	float		fDist, fAtten;
} PATH_INF_TYPE;

/***************************************************************************************
							GLOBAL VARIABLES
***************************************************************************************/
RAY_SETTINGS	gsRS;
RAY_TYPE*		gpsRayStack;	/* always points to the top of the ray stack */
long			giNumRayStack;		/* number of rays in the stack */
long			giNumRaysTraced;	/* total number of rays traced */
int				giNumAbort;			/* total number of rays aborted due to topological */ 
int				giRayMax;			/* size of the ray stack */
										/* or floating point errors */
long			giNumInter;		/* track the number of degenerate intersections during */
										/* ray walking */
long			giNumSavedPaths;

double			gfArea;
int				giNA;
long			giNumWalked;

long			giInitRayNum;		/* number of initial rays created */
FILE*			gfpRay;				/* for saving RaySave.dat */
PLANE_TYPE		gsGroundPlane;
PLANE_TYPE		gsCeilingPlane;
TRAP_TYPE		gsTrap;
VERTEX_TYPE 	(*gasTriVect)[3];
FILE*			gfpSave_Paths;
						  
POINT_TYPE	Point_Offset(POINT_TYPE sP1, POINT_TYPE sP2);
float Compute_fZ_For_Intersection(POINT_TYPE sInter, PARA_LINE_TYPE sL);
void Print_Ray(RAY_TYPE sRay);
void fPrint_Ray(FILE* fp, RAY_TYPE sRay);
long From_Edge(RAY_TYPE sRay, POINT_TYPE sPIn);
void Print_Tetra(TRAP_TYPE sT);
float T_Dist(VECTOR_TYPE sV, POINT_TYPE sP, POINT_TYPE sP2);
void Print_Ray_Tri(RAY_TRI_TYPE asTri[], short iNum);
void Check_Ray_Cone(RAY_TYPE sRay, RAY_TYPE* psRay);
void Extend_Points(POINT_TYPE* psP1, POINT_TYPE* psP2);
float Triangle_Area(POINT_TYPE asTriP[]);
int	All_In_Triangle(POINT_TYPE asP[], POINT_TYPE asP2[], RAY_TYPE sRay, PLANE_TYPE sPlane);

#endif
